@@ -21,7 +21,7 @@ import logging
 import ftputil
 from ftputil.error import FTPOSError
 import colemen_string_utils as strUtils
-import utils.file
+import utils.file as file
 import utils.objectUtils as obj
 import utils.file_write as write
 # import utils.file as f
@@ -29,7 +29,10 @@ import utils.resources
 
 
 logger = logging.getLogger(__name__)
-get_files = file.get_files
+
+
+def get_files(search_path=False, **kwargs):
+    file.get_files(search_path, **kwargs)
 
 
 def create(path, dir_name=False, **kwargs):
